/*
 *
 *  Linux Basic Security Modules
 *  LBSM project team. <lbsm@sec.jri.co.jp>
 *
 *  Copyright (c) 2002-2004 The Japan Research Institute, Limited
 *
 *  This is a part of Linux Basic Security Modules.
 *
 *  This programme is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This programme is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free software Foundation, Inc.,
 *  59 Temple Place, suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <linux/types.h>
#include <errno.h>
#include <linux/mm.h>
#include <linux/spinlock.h>

#ifndef _LBSM_HASHTAB_H
#define _LBSM_HASHTAB_H

#ifdef __cplusplus
extern "C" {
#endif

/**
 *  Basic data definition. In the hash table, we use character as 
 *  hash key value.
 */
typedef char *lbsm_hashTable_key_t;	/* generic key type  */
typedef void *lbsm_hashTable_data_t;	/* generic data type */

/**
 *  This structure defines data structure for hash Table's each node.
 *  This has three values. The "key" is hash key. The "data" is data
 *  includes hash table. The "next" is pointer for creating link structure.
 *  In the hash table, some nodes has same key are linked by using this 
 *  pointer.
 */
struct _lbsm_hashTable_node {
	lbsm_hashTable_key_t		key;	/* hash key                         */
	lbsm_hashTable_data_t		data;	/* data itself	                    */
	struct _lbsm_hashTable_node *	next;	/* next node pointer                */
};
typedef struct _lbsm_hashTable_node lbsm_hashTable_node_t;


struct _lbsm_hashTable {
        unsigned int  status;
	lbsm_hashTable_node_t **	htable; /* hash table                       */
	unsigned int 			size;   /* number of slots in hash table    */
	__u32 				e_num;  /* number of elements in hash table */
	spinlock_t			lock;   /* table locked                     */
	unsigned int (*hash_code) (struct _lbsm_hashTable *h, lbsm_hashTable_key_t key); 
						/* hash function                    */
	int (*equals) (struct _lbsm_hashTable *h, lbsm_hashTable_key_t key1, lbsm_hashTable_key_t key2); 
						/* key comparison function          */
        int (*destroy) (struct _lbsm_hashTable *htp, lbsm_hashTable_key_t k, lbsm_hashTable_data_t d);
						/* hash node delete function */
};
typedef struct _lbsm_hashTable lbsm_hashTable_t;

/**
 *  Define status codes for hash table functions 
 */
#define HASHTAB_SUCCESS     0
#define HASHTAB_OVERFLOW    -ENOMEM
#define HASHTAB_PRESENT     -EEXIST
#define HASHTAB_MISSING     -ENOENT

/**
 *  MACRO's for lock 
 */
#ifdef __KERNEL__
#define HTAB_LOCK(htab)   spin_lock_irqsave(&((htab)->lock), flags)
#define HTAB_UNLOCK(htab) spin_unlock_irqrestore(&((htab)->lock), flags)
#else
#define HTAB_LOCK(htab)
#define HTAB_UNLOCK(htab)
#endif

/**
 *  Functions
 */
#ifdef __STDC__

extern lbsm_hashTable_t *lbsm_hashTable_create(unsigned int size);
extern int lbsm_hashTable_set_hash_code(lbsm_hashTable_t *htp, 
				 unsigned int (*hash_code) (lbsm_hashTable_t *hth, lbsm_hashTable_key_t key));
extern int lbsm_hashTable_set_equals(lbsm_hashTable_t *htp, 
			      int (*equals) (lbsm_hashTable_t *htp,
					     lbsm_hashTable_key_t key1,
					     lbsm_hashTable_key_t key2));
extern int lbsm_hashTable_set_destroy(lbsm_hashTable_t *htp, 
			      int (*destroy) (lbsm_hashTable_t *htp,
					      lbsm_hashTable_key_t k,
					      lbsm_hashTable_data_t data));

extern int lbsm_hashTable_insert(lbsm_hashTable_t *h, lbsm_hashTable_key_t k, lbsm_hashTable_data_t d);
extern int lbsm_hashTable_remove(lbsm_hashTable_t *h, lbsm_hashTable_key_t k);
extern int lbsm_hashTable_replace(lbsm_hashTable_t *h, lbsm_hashTable_key_t k, lbsm_hashTable_data_t d);
extern lbsm_hashTable_data_t lbsm_hashTable_search(lbsm_hashTable_t *h, lbsm_hashTable_key_t k);
extern void lbsm_hashTable_destroy(lbsm_hashTable_t *h);
extern void lbsm_hashTable_hash_eval(lbsm_hashTable_t *h, char *tag);
extern void lbsm_hashTable_hash_print(lbsm_hashTable_t *, void (*)(lbsm_hashTable_data_t *));

#else

extern lbsm_hashTable_t *lbsm_hashTable_create();
extern int lbsm_hashTable_set_hash_code();
extern int lbsm_hashTable_set_equals();
extern int lbsm_hashTable_set_destroy();
extern int lbsm_hashTable_insert();
extern int lbsm_hashTable_remove();
extern int lbsm_hashTable_replace();
extern lbsm_hashTable_data_t lbsm_hashTable_search();
extern void lbsm_hashTable_destroy();
extern void lbsm_hashTable_hash_eval();
extern void lbsm_hashTable_hash_print();

#endif

#ifdef __cplusplus
}
#endif

#endif /* !_LBSM_HASHTAB_H */
